//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"context"
	"net/http"

	oidc "github.com/coreos/go-oidc"
	"golang.org/x/oauth2"
)

type OIDCSession struct {
	Nonce string
	token *oauth2.Token
}

func NewOIDCSession() (s *Session, err error) {
	os := &OIDCSession{}
	if os.Nonce, err = generateRandomString(16); err != nil {
		return
	}

	if s, err = NewSession(); err != nil {
		return
	}
	s.oidc = os
	if err = auth.sessions.insert(s); err != nil {
		return nil, err
	}
	return
}

type OIDCBackend struct {
	issuerURL    string
	provider     *oidc.Provider
	verifier     *oidc.IDTokenVerifier
	oauth2Config oauth2.Config
}

func NewOIDCBackend(cfg *OIDCConfig) (b *OIDCBackend, err error) {
	// TODO: make ctx a parameter?
	ctx := context.Background()
	b = &OIDCBackend{issuerURL: cfg.IssuerURL}
	if b.provider, err = oidc.NewProvider(ctx, cfg.IssuerURL); err != nil {
		return
	}

	oidcConfig := &oidc.Config{
		ClientID: cfg.ClientID,
	}
	b.verifier = b.provider.Verifier(oidcConfig)

	b.oauth2Config = oauth2.Config{
		ClientID:     cfg.ClientID,
		ClientSecret: cfg.ClientSecret,
		Endpoint:     b.provider.Endpoint(),
		RedirectURL:  cfg.CallbackURL,
		Scopes:       []string{oidc.ScopeOpenID, "username", "aura_shows"},
	}

	return
}

func (b *OIDCBackend) String() string {
	return "OpenID Connect using Identity Provider: " + b.issuerURL
}

// TODO: implement session refresh go-routine
//       see: https://github.com/golang/oauth2/issues/84

func (b *OIDCBackend) CallbackHandler() http.Handler {
	return &oidcCallbackHandler{backend: b}
}

func (b *OIDCBackend) LoginSession(sID string, s *Session, w http.ResponseWriter, r *http.Request) {
	if s.State != SessionStateNew && s.State != SessionStateLoginFailed {
		sendHTTPErrorResponse(w, http.StatusConflict, "this session is already logged in or login is in progress")
		return
	}
	if s.oidc == nil {
		sendHTTPErrorResponse(w, http.StatusConflict, "this is not an OIDC session")
		return
	}

	if err := auth.sessions.updateState(sID, SessionStateLoginInProgress); err != nil {
		auth.sessions.remove(sID)
		sendHTTPErrorResponse(w, http.StatusInternalServerError, "Updating session state failed: "+err.Error())
		return
	}
	// TODO start login timeout go-routine
	http.Redirect(w, r, b.oauth2Config.AuthCodeURL(sID, oidc.Nonce(s.oidc.Nonce)), http.StatusFound)
}

type oidcCallbackHandler struct {
	backend *OIDCBackend
}

func (h *oidcCallbackHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	sID := r.URL.Query().Get("state")
	s := auth.sessions.get(sID)
	if s == nil {
		sendHTTPInvalidSessionResponse(w)
		return
	}

	if s.oidc == nil {
		sendHTTPErrorResponse(w, http.StatusConflict, "this is not an OIDC session")
		return
	}
	if s.State != SessionStateLoginInProgress {
		sendHTTPErrorResponse(w, http.StatusConflict, "this session is not in state login-in-progress")
		return
	}

	oauth2Token, err := h.backend.oauth2Config.Exchange(r.Context(), r.URL.Query().Get("code"))
	if err != nil {
		auth.sessions.remove(sID)
		sendHTTPErrorResponse(w, http.StatusBadRequest, "OAuth2 token exchange failed: "+err.Error())
		return
	}
	rawIDToken, ok := oauth2Token.Extra("id_token").(string)
	if !ok {
		auth.sessions.remove(sID)
		sendHTTPErrorResponse(w, http.StatusBadRequest, "OIDC verification failed: no id_token field in oauth2 token.")
		return
	}

	// Verify the ID Token signature and nonce.
	idToken, err := h.backend.verifier.Verify(r.Context(), rawIDToken)
	if err != nil {
		auth.sessions.remove(sID)
		sendHTTPErrorResponse(w, http.StatusInternalServerError, "OAuth2 ID token verification failed: "+err.Error())
		return
	}
	if idToken.Nonce != s.oidc.Nonce {
		auth.sessions.remove(sID)
		sendHTTPErrorResponse(w, http.StatusInternalServerError, "OAuth2 ID token verification failed: invalid nonce")
		return
	}

	userInfo, err := h.backend.provider.UserInfo(r.Context(), oauth2.StaticTokenSource(oauth2Token))
	if err != nil {
		auth.sessions.remove(sID)
		sendHTTPErrorResponse(w, http.StatusInternalServerError, "fetching OIDC UserInfo failed: "+err.Error())
		return
	}

	newS := &Session{}
	if err := userInfo.Claims(newS); err != nil {
		sendHTTPErrorResponse(w, http.StatusInternalServerError, "parsing OIDC UserInfo failed: "+err.Error())
		return
	}
	newS.State = SessionStateLoggedIn
	newS.oidc = &OIDCSession{Nonce: s.oidc.Nonce}
	newS.oidc.token = oauth2Token
	if err = auth.sessions.update(sID, newS); err != nil {
		auth.sessions.remove(sID)
		sendHTTPErrorResponse(w, http.StatusInternalServerError, "updating session failed: "+err.Error())
		return
	}
	sendHTTPResponse(w, http.StatusOK, "You are now logged in as: "+newS.Username)
}
